import docker
client = docker.from_env()

def pull_image(image):
    response = client.api.pull(image, stream=True, decode=True)
    for line in response:
        print(line['status'])

def build_container(image, local_dir, env):
    return client.api.create_container(
        image, 
        command=["tail", "-f", "/dev/null"],
        working_dir='/opt/compliance',
        environment=env,
        host_config=client.api.create_host_config(
            binds={
                local_dir: '/opt/compliance'
            }
        )
    )

def execute_instruction(container, instruction):
    exec_id = client.api.exec_create(container, cmd=["bash", "-c", instruction])
    output = client.api.exec_start(exec_id)
    inspect = client.api.exec_inspect(exec_id)
    return [inspect['ExitCode'], instruction, output]

def execute_steps(image, local_dir, steps, env):
    container = build_container(image, local_dir, env)
    try:
        client.api.start(container=container)

        total_result = []
        for i in steps:
            execute_result = execute_instruction(container, i)
            total_result.append(execute_result)
            if execute_result[0] != 0:
                return (False, total_result)
        
        return (True, total_result)
    finally:
        client.api.kill(container=container)
    



import argparse
import sys
import yaml
import os
import glob
import json
from schema import Schema, SchemaError, Or
from textwrap import wrap
from docker_runner import execute_steps
from git import Repo
from tempfile import TemporaryDirectory

STATUS_CHARACTERS = {
    "pass": ' (pass)',
    "warning": ' * (warning)',
    "security": ' !! (security)',
    "fail": ' ! (fail)',
    "info": ' i (info)'
}

RULE_SCHEMA = Schema({
    "name": str,
    "image": str,
    "tags": [str],
    "steps": [str],
    "type": Or("warning", "security", "fail", "info")
}, ignore_extra_keys=True)

def intersection(lst1, lst2):
    lst2 = [name.upper() for name in lst2]
    return [value for value in lst1 if value.upper() in lst2]

def verify_single_rule(args, rule):
    if len(intersection(rule['tags'], args.tags)) > 0:
        _print(args, rule['name'].ljust(50, '.'), end='')
        
        env = {
            'COMPLIANCE_RULE_NAME': rule['name'],
            'COMPLIANCE_RULE_TYPE': rule['type'],
            'COMPLIANCE_TAGS': ",".join(args.tags)
        }

        for e in args.env:
            e = e.split('=')
            env[e[0]] = e[1]

        output_result = execute_steps(rule['image'], args.target, rule['steps'], env)
        
        _verbose(args, '\n')
        _print_result(args, output_result)

        if output_result[0]:
            _print(args, STATUS_CHARACTERS["pass"])
            _json_result(args, rule['name'], output_result, "pass")
            return "pass"
        else:
            _print(args, STATUS_CHARACTERS[rule['type']])
            _print_fail_only(args, output_result)
            _json_result(args, rule['name'], output_result, rule['type'])
            return rule['type']

    else:
        _verbose(args, "No matching tags, skipping rule file")
        return None

json_output = {
    "checks": []
}
def _json_print(args, results):
    if args.output == 'json':
        json_output['target'] = args.target
        json_output['results'] = results
        print(json.dumps(json_output, indent=4))

def _json_result(args, rule_name, output, status):
    if args.output == 'json':
        json_output["checks"].append({
            "status": status,
            "ruleName": rule_name,
            "steps": [{
                "exitCode": i[0],
                "step": i[1],
                "output": i[2].decode("utf-8")
            } for i in output[1]]
        })

def _print_docker_output(args, result):
    for line_exec in result[1]:
        exit_message = None
        if line_exec[0] == 0:
            exit_message =  " (Exit Code: 0)"
        else:
            exit_message = " (Exit Code: "+ line_exec[0] + ")"

        _print(args, "\t" + line_exec[1] + exit_message)
        for line in line_exec[2].splitlines():
            _print(args, '\t\t'+ line.decode("utf-8"))

def _print_result(args, result):
    if args.verbose:
        _print_docker_output(args, result)

def _print_fail_only(args, result):
    if (not args.verbose) and (not args.hide_fail):
        _print_docker_output(args, result)

def _verbose(args, message):
    if args.verbose:
        print(message)

def _print(args, message, end="\n"):
    if args.output == None:
        print(message, end=end)

def is_git_target(target):
    return target.startswith('http') or target.startswith('git') or target.startswith('ssh')

def run_compliance(args):
    args.tags.append('all')

    _print(args, "Checking compliance for " + args.target)

    results = {
        'warning': 0,
        'security': 0,
        'fail': 0,
        'pass': 0,
        'info': 0,
        'total_tests': 0
    }

    with TemporaryDirectory() as temp_dir:
        if is_git_target(args.target):
            _print(args, "Target is a git URL- cloning repository")
            _verbose(args, "Checking out git repo to " + temp_dir)
            _print(args, "Cloning git repo".ljust(50, '.'), end='')
            Repo.clone_from(args.target, 
                temp_dir, 
                multi_options=["--depth=1"], 
                branch=args.branch)
            _print(args, STATUS_CHARACTERS["pass"])
            args.target = temp_dir

        rule_files = glob.glob(args.rules, recursive=True)
        for filename in rule_files:
            _verbose(args, "Using rule file "+ filename)
            with open(filename, 'r') as stream:
                rule = yaml.safe_load(stream)
                try:
                    RULE_SCHEMA.validate(rule)
                except SchemaError as se:
                    print("Rule file {file} is invalid!".format(file=filename))
                    raise se
                response = verify_single_rule(args, rule)
                if response:
                    results[response] += 1
                    results['total_tests'] += 1
        
    _print(args, "{total_tests} tests ({pass} pass, {info} info, {warning} warning, {fail} fail, {security} security)".format(**results))
    _json_print(args, results)

    if args.no_error:
        sys.exit(0)

    if results['security'] > 0 or results['fail'] > 0:
        sys.exit(1)
    else:
        sys.exit(0)

def list_rules(args):
    rule_files = glob.glob(args.rules, recursive=True)
    for filename in rule_files:
        with open(filename, 'r') as stream:
            rule = yaml.safe_load(stream)
            try:
                RULE_SCHEMA.validate(rule)
            except SchemaError as se:
                print("Rule file {file} is invalid!".format(file=filename))
                raise se
            print("{name} ({type}) ".format(
                name=rule['name'], 
                type=rule['type']
            ).ljust(80, '-'))
            print("""File:\t{filename}
Tags:\t{tags}
Description:
{description}
""".format(
                filename=filename,
                description="\n".join(wrap(rule['description'], subsequent_indent='\t', initial_indent='\t')),
                tags=",".join(rule['tags'])))


def main():
    parser = argparse.ArgumentParser(description="Run compliance checks against a project")
    parser.set_defaults(func=lambda args: parser.print_help())
    subparsers = parser.add_subparsers()

    parser_run = subparsers.add_parser('run')
    parser_run.add_argument("-r", "--rules", help="glob of rule files", default="rules/**/*.yml")
    parser_run.add_argument("-v", "--verbose", help="increase output verbosity", action='store_true')
    parser_run.add_argument("-f", "--hide-fail", help="hide failure out information", action='store_true')
    parser_run.add_argument("--no-error", help="dont exit 1 if security and fails are found", action='store_true')
    parser_run.add_argument("-t", "--tags", help="project tags", action='append', default=[])
    parser_run.add_argument("-e", "--env", help="environmental variables", action='append', default=[])
    parser_run.add_argument("--output", help="output type", choices=["json"])
    parser_run.add_argument("-b", "--branch", help="the branch to use if target is a git repo", default="master")
    parser_run.add_argument("target", help="the target folder or git repo to run the compliance check on")
    parser_run.set_defaults(func=run_compliance)

    parser_list = subparsers.add_parser('list')
    parser_list.add_argument("-r", "--rules", help="glob of rule files", default="rules/**/*.yml")
    parser_list.set_defaults(func=list_rules)

    args = parser.parse_args()
    args.func(args)

if __name__ == "__main__":
    main()




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cmply', 'docker_runner']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.29,<4.0.0',
 'docker>5',
 'pyyaml>=6.0,<7.0',
 'schema>=0.7.5,<0.8.0']

entry_points = \
{'console_scripts': ['cmply = cmply:main']}

setup_kwargs = {
    'name': 'cmply',
    'version': '0.1.4',
    'description': 'A simple tool to validate if a repository adhears to specified rules.',
    'long_description': '# cmply\n\n[![PyPI version](https://img.shields.io/pypi/v/cmply.svg?logo=pypi&logoColor=FFE873)](https://pypi.org/project/cmply)\n[![Supported Python versions](https://img.shields.io/pypi/pyversions/cmply.svg?logo=python&logoColor=FFE873)](https://pypi.org/project/cmply)\n[![PyPI downloads](https://img.shields.io/pypi/dm/cmply.svg)](https://pypistats.org/packages/cmply)\n[![Licence](https://img.shields.io/github/license/pselibas/cmply.svg)](LICENSE.txt)\n\n\n## Installation\n\n```bash\npython3 -m pip install --upgrade cmply\n```\n\n## Introduction\n\nCmply is a tool used to run a set of rules against a repository.\nThe output will show the result of what rules fail or pass.\nThere is provision for different levels of failures. These levels are `info, warning, fail and security`.\nThe target can be a local folder or a git repository. \nIf a git repository is specified, the local git is used with any configure authentication relative to the local system.\n\n## Usage\n```bash\nusage: cmply run [-h] [-r RULES] [-v] [-f] [--no-error] [-t TAGS] [-e ENV] [--output {json}] [-b BRANCH] target\n\npositional arguments:\n  target                the target folder or git repo to run the compliance check on\n\noptions:\n  -h, --help            show this help message and exit\n  -r RULES, --rules RULES\n                        glob of rule files\n  -v, --verbose         increase output verbosity\n  -f, --hide-fail       hide failure out information\n  --no-error            dont exit 1 if security and fails are found\n  -t TAGS, --tags TAGS  project tags\n  -e ENV, --env ENV     environmental variables\n  --output {json}       output type\n  -b BRANCH, --branch BRANCH\n                        the branch to use if target is a git repo\n```\n\n## Rule file\n\nThe rules files are YAML files with the following format:\n```yaml\nname: sonarcube-setup   # Name your rule here\ndescription: |          # This is a verbose description\n  All projects must be registered in sonarcube\nimage: node:16          # Your steps are executed in this docker container\ntype: fail              # The type of rule info, warning, fail or security\ntags:                   # A list of tags. These then matched to the tags passed in from the command line for your execution.\n  - all\nsteps:                  # A list of commands executed. When a command has a non 0 response- the test is considered failed\n  - test -f sonar-project.properties\n```\n\n## Run using docker\n\nTo use the docker image of cmply it is important to pass through the hosts docker socket.\n\n```bash\ndocker run \\\n    -v //var/run/docker.sock:/var/run/docker.sock\n    -v ./rules:/opt/rules\n    -v ./target:/opt/target\n    -ti pselibas/cmply \\\n    run -r /opt/rules \\\n    /opt/target\n```\n\n',
    'author': 'Paul Selibas',
    'author_email': 'pselibas@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/pselibas/cmply',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4',
}


setup(**setup_kwargs)
